from typing import List
from thinkgear.data_points import DATA_POINTS, DataPointType, RawDataPoint


EXCODE: int = 0x55


def _create_data_point(level: int, code: int, data: bytes) -> DataPointType:
    point_type = DATA_POINTS.get(code, RawDataPoint)
    if len(data) < point_type.SIZE:
        point_type = RawDataPoint
    return point_type(level, code, data)


def parse(payload: bytes) -> List[DataPointType]:
    """Parse packet from ThinkGear Serial Stream.
    http://wearcam.org/ece516/mindset_communications_protocol.pdf"""
    pLength = len(payload)
    bytesParsed = 0
    code = 0
    length = 0
    extendedCodeLevel = 0
    data_points: List[DataPointType] = []

    # Loop until all bytes are parsed from the payload[] array...
    while bytesParsed < pLength:
        # Parse the extendedCodeLevel, code, and length
        extendedCodeLevel = 0
        while payload[bytesParsed] == EXCODE:
            extendedCodeLevel += 1
            bytesParsed += 1
        code = payload[bytesParsed]
        bytesParsed += 1
        # ~ if code & 0x80 and code not in [0xBA, 0xBC]:
        if code in [0x80, 0x83]:
            length = payload[bytesParsed]
            bytesParsed += 1
        else:
            length = 1

        data = payload[bytesParsed : bytesParsed + length]
        data_points.append(_create_data_point(extendedCodeLevel, code, data))

        bytesParsed += length
    return data_points



from typing import Tuple, Optional, List

import bluetooth


from thinkgear.parser import parse
from thinkgear.data_points import DataPointType, RawDataPoint

START_OF_PACKET = b"\xaa\xaa"
START_OF_PACKET_SIZE = len(START_OF_PACKET)

BUFFER_SIZE = 4096


def discover(lookup_name: str = "MindWave") -> Optional[Tuple[str, int]]:
    """Find some device by name."""
    nearby_devices = bluetooth.discover_devices(lookup_names=True)
    for address, name in nearby_devices:
        if lookup_name in name:
            services = bluetooth.find_service(address=address)
            try:
                port = next(
                    i["port"]
                    for i in services
                    if i["protocol"] == "RFCOMM" and i["port"]
                )
                return (address, port)
            except StopIteration:
                pass
    return None


def _check_sum(payload: bytes, check_sum: int) -> bool:
    return (~(sum(payload) & 0xFF) & 0xFF) == check_sum


class ThinkGearProtocol:
    """Read data by ThinkGear Serial Stream Protocol."""

    def __init__(self) -> None:
        self._buffer: bytearray = bytearray()
        self._data_points: List[DataPointType] = []

    def _recv(self) -> bytes:
        """Receive from device."""
        return b""

    def read(self) -> Optional[DataPointType]:
        """Read one data point."""
        if len(self._data_points) <= 0:
            self._data_points = self.readall()
        if len(self._data_points) <= 0:
            return None
        return self._data_points.pop()

    def skip_to_beginning(self) -> None:
        """Find beginning of packet."""
        data_beginning = self._buffer.find(START_OF_PACKET)
        self._buffer = self._buffer[data_beginning:]

    def pop_packet(self, recv: bool = True) -> bytes:
        """Get data from one packet."""
        if recv:
            self._buffer += self._recv()

        while True:
            self.skip_to_beginning()

            if len(self._buffer) < 3:
                # there is not enough data in buffer, exit
                return b""

            size = self._buffer[2]
            if size <= 0:
                # got bad size
                del self._buffer[0]
                continue

            if size + 4 > len(self._buffer):
                # there is not enough data in buffer, exit
                return b""

            packet = self._buffer[3 : size + 3]
            check_sum = self._buffer[size + 3]
            del self._buffer[: size + 4]

            if not _check_sum(packet, check_sum):
                continue

            return packet

        return b""

    def pop(self, recv: bool = True) -> List[DataPointType]:
        return parse(self.pop_packet(recv))

    def readall(self) -> List[DataPointType]:
        """Read bunch of data points."""
        self._buffer += self._recv()

        data_points = self._data_points
        self._data_points = []
        points: List[DataPointType] = [RawDataPoint(0, 0, b"")]

        while points:
            points = self.pop(False)
            data_points += points

        if not data_points:
            raise BlockingIOError("No data points")

        return data_points


class ThinkGearBluetooth(ThinkGearProtocol):
    """Bluetooth communication with Think Gear device."""

    def __init__(self, address: Optional[Tuple[str, int]] = None):
        super().__init__()
        self._address: Optional[Tuple[str, int]] = address
        self.socket: Optional[bluetooth.BluetoothSocket] = None

    @staticmethod
    def connect_device(
        address: Optional[Tuple[str, int]] = None
    ) -> Optional[bluetooth.BluetoothSocket]:
        """Connect device using address and port."""
        if address is None:
            address = discover()

        if address is None:
            return None

        socket = bluetooth.BluetoothSocket(bluetooth.RFCOMM)
        try:
            socket.connect(address)
        except bluetooth.btcommon.BluetoothError:
            return None
        return socket

    def connect(self) -> None:
        """Connect to device."""
        self.socket = self.connect_device(self._address)

    def is_connected(self) -> bool:
        """Check is connected to device."""
        return self.socket is not None

    def _recv(self) -> bytes:
        """Receive data from bluetooth device."""
        if self.socket is None:
            return b""
        return self.socket.recv(BUFFER_SIZE)


ThinkGear = ThinkGearBluetooth



from typing import List, Dict, Union, Type
from collections import namedtuple


def _eeg_from_bytes(data: bytes) -> List[int]:
    """Convert series of 3-byte unsigned integers to list of integers."""
    return [
        int.from_bytes(data[i : i + 3], byteorder="big")
        for i in range(int(len(data) / 3))
    ]


class DataPoint(namedtuple("DataPoint", "level, code, data, value")):
    """Data point, contains raw data bytes and integer value from this data."""

    SIZE = 1

    def __new__(cls, level: int, code: int, data: bytes) -> "DataPoint":
        return super().__new__(cls, level, code, data, data[0])


class UnknownDataPoint(DataPoint):
    """Unknown data point."""

    def __str__(self) -> str:
        return f"Unknown OpCode. Value: {self.value}"


class BatteryDataPoint(DataPoint):
    """Battery level data point."""

    def __str__(self) -> str:
        return f"Battery Level: {self.value}"


class PoorSignalDataPoint(DataPoint):
    """Poor signal level data point."""

    @property
    def has_contact(self) -> bool:
        """Headset has contact to skin."""
        return bool(self.value < 200)

    def __str__(self) -> str:
        string = f"Poor Signal Level: {self.value}"
        if not self.has_contact:
            string += " - NO CONTACT TO SKIN"
        return string


class AttentionDataPoint(DataPoint):
    """Attention level data point."""

    def __str__(self) -> str:
        return f"Attention Level: {self.value}"


class MeditationDataPoint(DataPoint):
    """Meditation level data point."""

    def __str__(self) -> str:
        return f"Meditation Level: {self.value}"


class BlinkDataPoint(DataPoint):
    """Blink level data point."""

    def __str__(self) -> str:
        return f"Blink Level: {self.value}"


class RawDataPoint(namedtuple("RawDataPoint", "level, code, data, value")):
    """Raw data point."""

    SIZE = 0

    def __new__(cls, level: int, code: int, data: bytes) -> "RawDataPoint":
        return super().__new__(
            cls, level, code, data, int.from_bytes(data, byteorder="big", signed=True)
        )

    def __str__(self) -> str:
        return f"Raw Value: {self.value}"


class EegDataPoints(
    namedtuple(
        "RawDataPoint",
        "level, code, data, delta, theta, lowAlpha, highAlpha, lowBeta, highBeta, lowGamma, midGamma",
    )
):
    """Electroencephalogram data points."""

    SIZE = 8

    def __new__(cls, level: int, code: int, data: bytes) -> "EegDataPoints":
        return super().__new__(cls, level, code, data, *_eeg_from_bytes(data))

    def __str__(self) -> str:
        # pylint: disable=missing-format-attribute
        return f"""EEG Powers:
                delta: {self.delta}
                theta: {self.theta}
                lowAlpha: {self.lowAlpha}
                highAlpha: {self.highAlpha}
                lowBeta: {self.lowBeta}
                highBeta: {self.highBeta}
                lowGamma: {self.lowGamma}
                midGamma: {self.midGamma}
                """


DataPointType = Union[DataPoint, RawDataPoint, EegDataPoints]

DATA_POINTS: Dict[int, Type[DataPointType]] = {
    0x01: BatteryDataPoint,
    0x02: PoorSignalDataPoint,
    0x04: AttentionDataPoint,
    0x05: MeditationDataPoint,
    0x16: BlinkDataPoint,
    0x80: RawDataPoint,
    0x83: EegDataPoints,
    0xBA: UnknownDataPoint,
    0xBC: UnknownDataPoint,
}



import unittest
from thinkgear.data_points import (
    PoorSignalDataPoint,
    EegDataPoints,
    AttentionDataPoint,
    MeditationDataPoint,
    RawDataPoint,
)
from thinkgear.parser import parse
from thinkgear.think_gear import ThinkGearProtocol

DATASHEET_EXAMPLE = b"\xaa\xaa\x20\x02\x00\x83\x18\x00\x00\x94\x00\x00\x42\x00\x00\x0b\x00\x00\x64\x00\x00\x4d\x00\x00\x3d\x00\x00\x07\x00\x00\x05\x04\x0D\x05\x3d\x34"


class ThinkGearTest(ThinkGearProtocol):
    def __init__(self):
        super().__init__()
        self.recv_data = b""

    def _recv(self) -> bytes:
        return self.recv_data


class TestParserMethods(unittest.TestCase):
    def test_parse(self):
        data = parse(DATASHEET_EXAMPLE[3:-1])
        self.assertEqual(
            list(map(type, data)),
            [
                PoorSignalDataPoint,
                EegDataPoints,
                AttentionDataPoint,
                MeditationDataPoint,
            ],
        )
        data = parse(b"\x83\x33\x34\x55")
        self.assertEqual(list(map(type, data)), [RawDataPoint])
        data = parse(b"\x03")
        self.assertEqual(list(map(type, data)), [RawDataPoint])
        data = parse(b"\x02")
        self.assertEqual(list(map(type, data)), [RawDataPoint])


class TestThinkGearMethods(unittest.TestCase):
    def test_skip_to_beginning(self):
        tg = ThinkGearTest()
        tg._buffer = bytearray(b"\xaa\xaa")
        tg.skip_to_beginning()
        self.assertEqual(tg._buffer, b"\xaa\xaa")
        tg._buffer = bytearray(b"1231\xaa\xaa")
        tg.skip_to_beginning()
        self.assertEqual(tg._buffer, b"\xaa\xaa")
        tg._buffer = bytearray(b"1231\xaa\xaa123")
        tg.skip_to_beginning()
        self.assertEqual(tg._buffer, b"\xaa\xaa123")
        tg._buffer = bytearray(b"")
        tg.skip_to_beginning()
        self.assertEqual(tg._buffer, b"")
        tg._buffer = bytearray(b"\x00")
        tg.skip_to_beginning()
        self.assertEqual(tg._buffer, b"\x00")
        tg._buffer = bytearray(b"\x00\x01")
        tg.skip_to_beginning()
        self.assertEqual(tg._buffer, b"\x01")

    def test_pop_packet(self):
        tg = ThinkGearTest()
        tg.recv_data = b""
        self.assertEqual(tg.pop_packet(), b"")
        tg.recv_data = b"\x00"
        self.assertEqual(tg.pop_packet(), b"")
        tg.recv_data = b"\x00\xaa\xaa"
        self.assertEqual(tg.pop_packet(), b"")
        tg._buffer = bytearray()
        tg.recv_data = b"\xaa\xaa\x00\xaa\xaa\x01\x34\xcb"
        self.assertEqual(tg.pop_packet(), b"\x34")
        tg.recv_data = b"\xaa\xaa\x01\x23\x00\xaa\xaa\x01\x34\xcb"
        self.assertEqual(tg.pop_packet(), b"\x34")
        tg.recv_data = b"\xaa\xaa\x10"
        self.assertEqual(tg.pop_packet(), b"")
        tg._buffer = bytearray()
        tg.recv_data = b"\xaa\xaa\x02\x00\x00\xff"
        self.assertEqual(tg.pop_packet(), b"\x00\x00")
        tg.recv_data = DATASHEET_EXAMPLE[:]
        self.assertEqual(tg.pop_packet(), DATASHEET_EXAMPLE[3:-1])


if __name__ == "__main__":
    unittest.main()
